/*
 *
 * j-chkmail - Mail Server Filter for sendmail
 *
 * Copyright (c) 2001, 2002, 2003, 2004 - Ecole des Mines de Paris
 *
 *  Auteur     : Jose Marcio Martins da Cruz
 *               Jose-Marcio.Martins@ensmp.fr
 *
 *  Historique :
 *  Creation     : janvier 2002
 *
 * This program is free software, but with restricted license :
 *
 * - j-chkmail is distributed only to registered users
 * - j-chkmail license is available only non-commercial applications,
 *   this means, you can use j-chkmail if you make no profit with it.
 * - redistribution of j-chkmail in any way : binary, source in any
 *   media, is forbidden
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * More details about j-chkmail license can be found at j-chkmail
 * web site : http://j-chkmail.ensmp.fr
 */


#ifndef __MACROS_H__

#define ASSERT(a)		assert((a))

#define FREE(x) \
  do { \
    if (x != NULL) \
      free(x); \
    x = NULL; \
  } while (0)


#define  FD_PRINTF(fdp, ...)					    \
  do {								    \
    if (fdp >= 0) {						    \
      char fdp_str[1024];					    \
      (void ) snprintf(fdp_str, sizeof(fdp_str), __VA_ARGS__);	    \
      if (write (fdp, fdp_str, strlen(fdp_str)) != strlen(fdp_str)) \
        LOG_SYS_ERROR("error on FD_PRINTF");			    \
    } else							    \
      printf(__VA_ARGS__);					    \
  } while (0)

#define MUTEX_LOCK(mutex)                                    \
  {                                                          \
    int r = 0;                                               \
    if ((r = pthread_mutex_lock(mutex)) != 0) {              \
      LOG_SYS_ERROR("pthread_mutex_lock : %s", strerror(r)); \
    }                                                        \
  }

#define MUTEX_UNLOCK(mutex)                                    \
  {                                                            \
    int r = 0;                                                 \
    if ((r = pthread_mutex_unlock(mutex)) != 0) {              \
      LOG_SYS_ERROR("pthread_mutex_unlock : %s", strerror(r)); \
    }                                                          \
  }


#ifndef HAVE_MAX
#define max(a,b)                 ((a) > (b) ? (a) : (b))
#define min(a,b)                 ((a) < (b) ? (a) : (b))
#endif

#ifdef   MAX
# undef  MAX
#endif

#ifdef   MIN
# undef  MIN
#endif

#define MAX(a,b)              ((a) > (b) ? (a) : (b))
#define MIN(a,b)              ((a) < (b) ? (a) : (b))

#ifndef TRUE
# define TRUE   1
#endif             /* ! TRUE */
#ifndef FALSE
# define FALSE  0
#endif             /* ! FALSE */


#define CRLF                     "\r\n"

#define ADJUST_FILENAME(path, fname, cfdir, defval)		\
  do {									\
    char *name = NULL;							\
    name = (fname != NULL && strlen(fname) > 0) ? fname : defval;	\
    strcpy(path, "");							\
    if (*name == '/')							\
      strlcpy(path, name, sizeof(path));				\
    else								\
      snprintf(path, sizeof(path), "%s/%s", cfdir, name);		\
    MESSAGE_INFO(15, "Adjusted path is %s", path);			\
  } while (FALSE);

#define __MACROS_H__
#endif

