/*
 *
 * sympa-milter.c
 *
 *
 */
#include <j-sys.h>
#if 1
#include <application.h>
#endif
#include "libmilter/mfapi.h"

#define CRLF   "\r\n"
#define LF     "\n"

#if EOL_IS_CRLF
# define LINESEP   CRLF
#else
# define LINESEP   LF
#endif

#define HDRSEP     ": "


/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static sfsistat     mlfi_connect(SMFICTX *, char *, _SOCK_ADDR *);
static sfsistat     mlfi_helo(SMFICTX *, char *);
static sfsistat     mlfi_envfrom(SMFICTX *, char **);
static sfsistat     mlfi_envto(SMFICTX *, char **);
static sfsistat     mlfi_header(SMFICTX *, char *, char *);
static sfsistat     mlfi_eoh(SMFICTX *);
static sfsistat     mlfi_body(SMFICTX *, unsigned char *, size_t);
static sfsistat     mlfi_eom(SMFICTX *);
static sfsistat     mlfi_close(SMFICTX *);
static sfsistat     mlfi_abort(SMFICTX *);
static sfsistat     mlfi_cleanup(SMFICTX *, bool);


/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
typedef struct
{
  int                 fd;
  char                envto[256];

  char                rcpt[256];
  char                family[256];
  char                hdrto[256];

  char                fname[256];
} ok_autolist_T;

#define MAX_LRCPT   100

typedef struct
{
  int                 nl;
  ok_autolist_T       lists[MAX_LRCPT];
} mlfiPriv_T;


typedef struct
{
  char                family[256];
  char                rcpt_expr[256];
}
cf_autolist_T;


/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
#define MLFIPRIV(ctx)  (ctx != NULL ? (mlfiPriv_T *) smfi_getpriv(ctx) : NULL)

#define DEFAULT_CONF         (CONFDIR "/sympa-milter.conf")

#if 1
#define DEFAULT_SOCKET       "inet:2040@localhost"
#else
#define DEFAULT_SOCKET       ("local:" STATEDIR "/sympa-milter.sock")
#endif

#define DEFAULT_PIDFILE      (STATEDIR "/sympa-milter.pid")

#define DEFAULT_SPOOL_DIR    (STATEDIR)

static char        *sympa_conf = DEFAULT_CONF;

static char        *sympa_socket = NULL;
static char        *sympa_pid = NULL;

static char        *sympa_spool_dir = NULL;

static char        *sympa_user = NULL;
static char        *sympa_group = NULL;

static int          sympa_log_level = 9;
static char        *sympa_log_facility = "local5";
static bool         sympa_log_severity = FALSE;

static bool         sympa_include_domain = TRUE;

static void         dump_sympa_rcpt_table();
static bool         load_sympa_rcpt_table(char *fname);
static bool         check_sympa_rcpt(char *rcpt, ok_autolist_T * liste);

static bool         read_config(char *fname);

static bool         create_pid_file(void);
static void         remove_pid_file(void);

static void         usage(char *);
int                 daemon_init();

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              sfsistat
mlfi_connect(ctx, hostname, hostaddr)
     SMFICTX            *ctx;
     char               *hostname;
     _SOCK_ADDR         *hostaddr;
{
  mlfiPriv_T         *priv = NULL;

  if (hostname == NULL)
    hostname = "HOST_UNKNOWN";

  MESSAGE_INFO(10, "Connect from %s", hostname);

  if ((priv = malloc(sizeof (mlfiPriv_T))) == NULL)
  {
    LOG_SYS_ERROR("malloc error");
    return SMFIS_TEMPFAIL;
  }

  memset(priv, 0, sizeof (mlfiPriv_T));
  smfi_setpriv(ctx, priv);

  return SMFIS_CONTINUE;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              sfsistat
mlfi_helo(ctx, name)
     SMFICTX            *ctx;
     char               *name;
{
  return SMFIS_CONTINUE;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              sfsistat
mlfi_envfrom(SMFICTX * ctx, char **from)
{
  mlfiPriv_T         *priv = NULL;

  ASSERT(ctx != NULL);
  priv = MLFIPRIV(ctx);
  ASSERT(priv != NULL);

  mlfi_cleanup(ctx, FALSE);

  MESSAGE_INFO(11, "Message from %s", from[0]);

  return SMFIS_CONTINUE;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              sfsistat
mlfi_envto(SMFICTX * ctx, char **to)
{
  mlfiPriv_T         *priv = NULL;
  ok_autolist_T       liste;

  ASSERT(ctx != NULL);
  priv = MLFIPRIV(ctx);
  ASSERT(priv != NULL);

  ASSERT(to != NULL);
  ASSERT(to[0] != NULL);

  memset(&liste, 0, sizeof (liste));
  if (check_sympa_rcpt(to[0], &liste))
  {
    char                fname[256];
    char                buf[256];
    int                 fd;

    if (priv->nl >= MAX_LRCPT)
    {
      MESSAGE_WARNING(9, "Too many lists...");
      return SMFIS_REJECT;
    }

    MESSAGE_INFO(10, "Message to family=%s - recipient=%s", liste.family, to[0]);

    snprintf(fname, sizeof (fname), "%s/T.%s.%ld.XXXXXX", sympa_spool_dir,
             liste.family, time(NULL));

    fd = mkstemp(fname);

    if (fd < 0)
    {
      LOG_SYS_ERROR("can't create spool file (%s)", fname);
      fd = -1;
      return FALSE;
    }

    liste.fd = fd;
    strlcpy(liste.fname, fname, sizeof(fname));
    priv->lists[priv->nl++] = liste;

    snprintf(buf, sizeof (buf), "X-Sympa-Family: %s%s", liste.family, LINESEP);
    if (write(fd, buf, strlen(buf)) < strlen(buf))
      LOG_SYS_ERROR("Error writing to spool file");
    snprintf(buf, sizeof (buf), "X-Sympa-To: %s%s", liste.hdrto, LINESEP);
    if (write(fd, buf, strlen(buf)) < strlen(buf))
      LOG_SYS_ERROR("Error writing to spool file");

  }
  return SMFIS_CONTINUE;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */

static              sfsistat
mlfi_header(SMFICTX * ctx, char *headerf, char *headerv)
{
  mlfiPriv_T         *priv = NULL;
  int                 i;

  ASSERT(ctx != NULL);
  priv = MLFIPRIV(ctx);
  ASSERT(priv != NULL);

  if (headerf == NULL || strlen(headerf) == 0)
    return SMFIS_CONTINUE;

  if (headerv == NULL || strlen(headerv) == 0)
    return SMFIS_CONTINUE;

  for (i = 0; i < priv->nl; i++)
  {
    size_t              len;

    len = strlen(headerf);
    if (write(priv->lists[i].fd, headerf, len) < len)
      LOG_SYS_ERROR("Error writing to spool file");
    len = strlen(HDRSEP);
    if (write(priv->lists[i].fd, HDRSEP, len) < len)
      LOG_SYS_ERROR("Error writing to spool file");
    len = strlen(headerv);
    if (write(priv->lists[i].fd, headerv, len) < len)
      LOG_SYS_ERROR("Error writing to spool file");

    len = strlen(LINESEP);
    if (write(priv->lists[i].fd, LINESEP, len) < len)
      LOG_SYS_ERROR("Error writing to spool file");
  }

  return SMFIS_CONTINUE;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              sfsistat
mlfi_eoh(SMFICTX * ctx)
{
  mlfiPriv_T         *priv = NULL;
  int                 i;

  ASSERT(ctx != NULL);
  priv = MLFIPRIV(ctx);
  ASSERT(priv != NULL);

  for (i = 0; i < priv->nl; i++)
  {
    size_t              len;

    len = strlen(LINESEP);
    if (write(priv->lists[i].fd, LINESEP, len) < len)
      LOG_SYS_ERROR("Error writing to spool file");
  }

  return SMFIS_CONTINUE;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              sfsistat
mlfi_body(SMFICTX * ctx, unsigned char *buf, size_t size)
{
  mlfiPriv_T         *priv = NULL;
  int                 i;

  ASSERT(ctx != NULL);
  priv = MLFIPRIV(ctx);
  ASSERT(priv != NULL);

#if !EOL_IS_CRLF
  if (priv->nl > 0)
  {
    char               *p, *q;
    size_t              nsize;

    p = q = buf;
    nsize = 0;
    for (i = 0; i < size; i++, p++)
    {
      if (*p != '\r')
      {
        *q++ = *p;
        nsize++;
      }
    }
    *q = '\0';
    size = nsize;
  }
#endif

  for (i = 0; i < priv->nl; i++)
  {
    if (write(priv->lists[i].fd, buf, size) < size)
      LOG_SYS_ERROR("Error writing to spool file");
  }

  return SMFIS_CONTINUE;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              sfsistat
mlfi_eom(SMFICTX * ctx)
{
  mlfiPriv_T         *priv = NULL;
  int                 i;

  ASSERT(ctx != NULL);
  priv = MLFIPRIV(ctx);
  ASSERT(priv != NULL);

  for (i = 0; i < priv->nl; i++)
  {
    int                 r;
    char                fname[256];
    int                 ri;

    close(priv->lists[i].fd);

    ri = random() % 1024;
    snprintf(fname, sizeof (fname), "%s/%s.%ld.%04d", sympa_spool_dir,
             priv->lists[i].family, time(NULL), ri);
    if (rename(priv->lists[i].fname, fname) != 0)
    {
      LOG_SYS_ERROR("Can't rename %s", priv->lists[i].fname);
    }

    r = smfi_delrcpt(ctx, priv->lists[i].rcpt);
    if (r != 0)
      MESSAGE_INFO(9, "Removing %s : %d", priv->lists[i].rcpt, r);
  }

  mlfi_cleanup(ctx, FALSE);

  return SMFIS_CONTINUE;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              sfsistat
mlfi_close(SMFICTX * ctx)
{
  mlfiPriv_T         *priv = NULL;

  ASSERT(ctx != NULL);
  priv = MLFIPRIV(ctx);
  ASSERT(priv != NULL);

  mlfi_cleanup(ctx, TRUE);

  return SMFIS_CONTINUE;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              sfsistat
mlfi_abort(SMFICTX * ctx)
{
  mlfiPriv_T         *priv = NULL;
  int                 i;

  ASSERT(ctx != NULL);
  priv = MLFIPRIV(ctx);
  ASSERT(priv != NULL);

  for (i = 0; i < priv->nl; i++)
  {
    int                 r;

    if (priv->lists[i].fd > 0)
      close(priv->lists[i].fd);

    if (remove(priv->lists[i].fname) != 0)
    {
      LOG_SYS_ERROR("Can't remove %s", priv->lists[i].fname);
    }

  }

  mlfi_cleanup(ctx, FALSE);

  return SMFIS_CONTINUE;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              sfsistat
mlfi_cleanup(SMFICTX * ctx, bool ok)
{
  mlfiPriv_T         *priv = NULL;
  int                 i;

  ASSERT(ctx != NULL);
  priv = MLFIPRIV(ctx);
  ASSERT(priv != NULL);

  for (i = 0; i < priv->nl; i++)
    close(priv->lists[i].fd);
  memset(priv->lists, 0, sizeof (priv->lists));
  priv->nl = 0;

  if (ok)
  {
    FREE(priv);
    smfi_setpriv(ctx, NULL);
  }

  return SMFIS_CONTINUE;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
struct smfiDesc     smfilter = {
  "Sympa-Milter",               /* filter name */
  SMFI_VERSION,                 /* version code -- do not change */
  SMFIF_ADDHDRS | SMFIF_CHGHDRS | SMFIF_CHGBODY | SMFIF_ADDRCPT | SMFIF_DELRCPT,
  /* flags */
  mlfi_connect,                 /* connection info filter */
  mlfi_helo,                    /* SMTP HELO command filter */
  mlfi_envfrom,                 /* envelope sender filter */
  mlfi_envto,                   /* envelope recipient filter */
  mlfi_header,                  /* header filter */
  mlfi_eoh,                     /* end of header */
  mlfi_body,                    /* body block filter */
  mlfi_eom,                     /* end of message */
  mlfi_abort,                   /* message aborted */
  mlfi_close                    /* connection cleanup */
#if HAVE_XXFI_UNKNOWN
    , NULL                      /* unknown command */
#endif
#if HAVE_XXFI_DATA
    , NULL                      /* data */
#endif
#if HAVE_XXFI_NEGOTIATE
    , NULL                      /* negotiate */
#endif
#if HAVE_XXFI_SIGNAL
    , NULL                      /* signale */
#endif
};

int
main(argc, argv)
     int                 argc;
     char               *argv[];
{
  int                 res = 0;
  int                 i;

  const char         *args = "p:c:dl:hs:";
  char                c;

  bool                debug = FALSE;

  log_level = 9;
  set_log_output(TRUE, FALSE);

  while ((c = getopt(argc, argv, args)) != -1)
  {
    switch (c)
    {
      case 'h':                /* OK */
        usage(argv[0]);
        exit(0);
        break;

      case 'p':
        sympa_socket = optarg;
        break;

      case 's':
        sympa_spool_dir = optarg;
        break;

      case 'c':
        sympa_conf = optarg;
        break;

      case 'd':
        debug = TRUE;
        break;

      case 'l':
        log_level = atoi(optarg);
        break;

      default:
        (void) fprintf(stderr, "Error reading command line options : %c\n", c);
        exit(1);
    }
  }

  set_log_output(TRUE, debug);

  (void) read_config(sympa_conf);

  if (sympa_socket == NULL)
    sympa_socket = DEFAULT_SOCKET;
  if (sympa_pid == NULL)
    sympa_pid = DEFAULT_PIDFILE;
  if (sympa_spool_dir == NULL)
    sympa_spool_dir = DEFAULT_SPOOL_DIR;
  if (sympa_user == NULL)
    sympa_user = RUN_AS_USER;
  if (sympa_group == NULL)
    sympa_group = RUN_AS_GROUP;

  log_level = sympa_log_level;
  log_severity = sympa_log_severity;
  {
    int                 x;

    x = facility_value(sympa_log_facility);
    if (x >= 0)
      log_facility = x;
  }
  openlog(PACKAGE, LOG_PID | LOG_NOWAIT | LOG_NDELAY, log_facility);

  MESSAGE_INFO(8, " * Starting sympa-milter");

  (void) daemon_init();

  if (!load_sympa_rcpt_table(sympa_conf))
  {
    MESSAGE_WARNING(9, "Configuration file not found");
  }

  if (debug)
    dump_sympa_rcpt_table();

  (void) smfi_setconn(sympa_socket);

  if (smfi_register(smfilter) == MI_FAILURE)
  {
    fprintf(stderr, "smfi_register failed\n");
    exit(EX_UNAVAILABLE);
  }

  set_log_output(TRUE, debug);

  if (create_pid_file())
  {
    res = smfi_main();
    remove_pid_file();
  }

  MESSAGE_INFO(9, "Exiting sympa-milter");
  return res;
}



/*****************************************************************************
 *                                                                           * 
 *                                                                           *
 *****************************************************************************/

static j_table_T    htbl;

static pthread_mutex_t st_mutex = PTHREAD_MUTEX_INITIALIZER;


/*****************************************************************************
 *                                                                           * 
 *                                                                           *
 *****************************************************************************/
void
dump_sympa_rcpt_table()
{
  cf_autolist_T       p;

  printf("*** Sympa rcpt table : \n");
  if (j_table_get_first(&htbl, &p) == 0)
  {
    do
    {
      printf("   RCPT : %-20s %s\n", p.family, p.rcpt_expr);
    } while (j_table_get_next(&htbl, &p) == 0);
  }
  printf("***  \n");
}

/*****************************************************************************
 *                                                                           * 
 *                                                                           *
 *****************************************************************************/
static int
rcpt_comp(pa, pb)
     const void         *pa;
     const void         *pb;
{
  cf_autolist_T      *a = (cf_autolist_T *) pa;
  cf_autolist_T      *b = (cf_autolist_T *) pb;

  int                 r;

  r = strcasecmp(a->family, b->family);
  if (r == 0)
    r = strcasecmp(a->rcpt_expr, b->rcpt_expr);

  return r;
}

/*****************************************************************************
 *                                                                           * 
 *                                                                           *
 *****************************************************************************/
static int
add_sympa_rcpt_rec(vk, vv)
     void               *vk;
     void               *vv;
{
  char               *k = (char *) vk;
  char               *v = (char *) vv;
  cf_autolist_T       r;

  memset(&r, 0, sizeof (r));
  if (k == NULL || strlen(k) == 0)
    return 0;
  if (v == NULL || strlen(v) == 0)
    return 0;

  strlcpy(r.rcpt_expr, v, sizeof (r.rcpt_expr));
  strlcpy(r.family, k, sizeof (r.family));

  return j_table_add(&htbl, &r);
}

/*****************************************************************************
 *                                                                           * 
 *                                                                           *
 *****************************************************************************/
static              bool
read_it(path, tag)
     char               *path;
     char               *tag;
{
  int                 r;

  r = j_rd_text_file(path, RD_TWO_COLUMN, RD_NORMAL, tag, add_sympa_rcpt_rec);

  return r >= 0;
}

static              bool
load_sympa_rcpt_table(fname)
     char               *fname;
{
  int                 res = 0;
  static int          htbl_ok = FALSE;
  bool                result = FALSE;

  MUTEX_LOCK(&st_mutex);
  if (htbl_ok == FALSE)
  {
    memset(&htbl, 0, sizeof (htbl));
    res = j_table_init(&htbl, sizeof (cf_autolist_T), 256, rcpt_comp);
    if (res == 0)
      htbl_ok = TRUE;
  }
  if (res == 0)
    res = j_table_clear(&htbl);

  if (res == 0)
  {
    char                name[1024];
    char               *bname, *dname, *p;
    char                name_tag[1024];

    bname = dname = NULL;
    strlcpy(name, fname, sizeof (name));

    bname = name;
    p = strrchr(name, '/');
    if (p != NULL)
    {
      *p = '\0';
      bname = ++p;
      dname = name;
    } else
    {
      bname = name;
      dname = ".";
    }

    snprintf(name_tag, sizeof (name_tag), "%s:families", bname);
    result = read_conf_data_file(dname, name_tag, DEFAULT_CONF, read_it);
    if (result)
      j_table_sort(&htbl);
  }

  MUTEX_UNLOCK(&st_mutex);

  return result;
}


/*****************************************************************************
 *                                                                           * 
 *                                                                           *
 *****************************************************************************/
static              bool
check_sympa_rcpt(rcpt, liste)
     char               *rcpt;
     ok_autolist_T      *liste;
{
  cf_autolist_T       p;
  int                 res = FALSE;;

  ASSERT(liste != NULL);

  if ((rcpt == NULL) || (strlen(rcpt) == 0))
    return FALSE;

  memset(&p, 0, sizeof (p));
  /*strlcpy(p.rcpt, rcpt, sizeof (p.rcpt)); */

  extract_email_address(liste->rcpt, rcpt, sizeof (liste->rcpt));

  MUTEX_LOCK(&st_mutex);
  if (j_table_get_first(&htbl, &p) == 0)
  {
    do
    {
      if (strexpr(liste->rcpt, p.rcpt_expr, NULL, NULL, TRUE))
      {
        strlcpy(liste->family, p.family, sizeof (liste->family));
        strlcpy(liste->envto, rcpt, sizeof (liste->envto));
	strlcpy(liste->hdrto, liste->rcpt, sizeof (liste->hdrto));
	if (!sympa_include_domain) {
	  char *p = NULL;

	  p = strchr(liste->hdrto, '@');
	  if (p != NULL)
	    *p = '\0';
	}
        res = TRUE;
        break;
      }
    } while (j_table_get_next(&htbl, &p) == 0);
  }
  MUTEX_UNLOCK(&st_mutex);

  return res;
}

/*****************************************************************************
 *                                                                           * 
 *                                                                           *
 *****************************************************************************/
static int
read_conf_line(s, arg)
     char               *s;
     void               *arg;
{
  char               *sa = NULL;
  char               *sb = NULL;

  if (s == NULL)
    return 1;

  sa = sb = s;

  MESSAGE_INFO(12, "   Conf line : %s", s);

  sb += strcspn(sa, " \t");
  if (*sb != '\0')
  {
    *sb++ = '\0';
    if (*s != '\0')
      sb += strspn(sb, " \t");
  }

  if (strlen(sa) == 0)
    return 1;

  if (sb == NULL || strlen(sb) == 0)
    return 1;

  if (STRCASEEQUAL(sa, "socket"))
  {
    MESSAGE_INFO(9, " SM SOCKET       : %s", sb);
    sympa_socket = strdup(sb);
  }

  if (STRCASEEQUAL(sa, "pid_file"))
  {
    MESSAGE_INFO(9, " PID FILE        : %s", sb);
    sympa_pid = strdup(sb);
  }

  if (STRCASEEQUAL(sa, "spool_dir"))
  {
    MESSAGE_INFO(9, " SPOOL DIR       : %s", sb);
    sympa_spool_dir = strdup(sb);
  }

  if (STRCASEEQUAL(sa, "log_level"))
  {
    int                 v;

    MESSAGE_INFO(9, " LOG_LEVEL       : %s", sb);
    v = atoi(sb);
    if (v >= 0)
      sympa_log_level = v;
  }

  if (STRCASEEQUAL(sa, "log_facility"))
  {
    MESSAGE_INFO(9, " LOG_FACILITY    : %s", sb);
    strtolower(sb);
    sympa_log_facility = strdup(sb);
  }

  if (STRCASEEQUAL(sa, "log_severity"))
  {
    MESSAGE_INFO(9, " LOG_SEVERITY    : %s", sb);
    sympa_log_severity = STRCASEEQUAL(sb, "YES");
  }

  if (STRCASEEQUAL(sa, "run_as_user"))
  {
    MESSAGE_INFO(9, " USER            : %s", sb);
    sympa_user = strdup(sb);
  }

  if (STRCASEEQUAL(sa, "run_as_group"))
  {
    MESSAGE_INFO(9, " GROUP           : %s", sb);
    sympa_group = strdup(sb);
  }

  if (STRCASEEQUAL(sa, "preserve_domain"))
  {
    MESSAGE_INFO(9, " PRESERVE DOMAIN : %s", sb);
    sympa_include_domain = STRCASEEQUAL(sb, "YES");
  }

  return 0;
}


static              bool
read_config(fname)
     char               *fname;
{

  MESSAGE_INFO(10, "Reading configuration file : %s", fname);
  return j_rd_file(fname, "general", read_conf_line, NULL);
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */

static              bool
create_pid_file(void)
{
  FILE               *fpid;
  char                logbuf[1024];
  char               *pid_file = sympa_pid;

  if (pid_file == NULL || strlen(pid_file) == 0)
  {
    MESSAGE_INFO(0, "pid_file : NULL pointer");
    return FALSE;
  }

  if (access(pid_file, F_OK) == 0)
  {
    MESSAGE_ERROR(0, "PID_FILE %s exists. Is there another j-chkmail running ?",
                  pid_file);
    exit(EX_SOFTWARE);
  }

  if ((fpid = fopen(pid_file, "w")) != NULL)
  {
    fprintf(fpid, "%d\n", (int) getpid());
    fclose(fpid);
    if (chmod(pid_file, S_IRUSR | S_IRGRP | S_IROTH) != 0)
      LOG_SYS_ERROR("error changing pid file mode");

    return TRUE;
  }
  LOG_SYS_ERROR("Error opening %s file", pid_file);

  if (errno != EINTR)
    exit(EX_SOFTWARE);
}

/* ****************************************************************************
 *                                                                            * 
 *                                                                            *
 **************************************************************************** */
static void
remove_pid_file(void)
{
  char               *pid_file = sympa_pid;
  pid_t               pid = getpid();

  if (pid <= 0)
    return;

  if (pid_file == NULL || strlen(pid_file) == 0)
  {
    MESSAGE_WARNING(9, "pid_file : NULL pointer");
    return;
  }

  remove(pid_file);
}

/*****************************************************************************
 *                                                                           * 
 *                                                                           *
 *****************************************************************************/
static void
usage(appli)
     char               *appli;
{
  printf("Usage : %s options\n"
         "        -h  : help\n"
         "        -p  : socket\n"
         "              inet:2040@localhost\n"
         "              local:/var/sock\n"
         "              (default = %s)\n"
         "        -c  : configuration file (default = %s)\n"
         "        -l  : log level (default = 9)\n"
         "        -d  : set debug on (sympa-milter will run in foreground)\n"
         "        -s  : set sympa spool directory (%s)\n",
         appli, DEFAULT_SOCKET, DEFAULT_CONF, DEFAULT_SPOOL_DIR);
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static int
set_uid_gid(user, group)
     char               *user;
     char               *group;
{
  struct passwd      *pw = NULL;
  struct group       *gr = NULL;
  int                 res = 0;

  uid_t               uid = getuid();
  gid_t               gid = getgid();

  if (uid != 0)
  {
    LOG_MSG_WARNING("Only root can set user and group");
    return 0;
  }

  if ((gr = getgrnam(group)) != NULL)
  {
    LOG_MSG_DEBUG(20, "GID DE %s : %ld", group, (long) gr->gr_gid);
    if (gid == gr->gr_gid)
    {
      LOG_MSG_WARNING("sympa-milter is already running as group %s", group);
    } else
    {
      if ((uid == 0) && (setregid(gr->gr_gid, gr->gr_gid) < 0))
      {
        LOG_SYS_ERROR("Can't set process gid = %ld", (long) gr->gr_gid);
        return 1;
      }
    }
  } else
  {
    LOG_SYS_ERROR("Error getgrnam(%s)", group);
    return 1;
  }

  if ((pw = getpwnam(user)) != NULL)
  {
    LOG_MSG_DEBUG(20, "UID DE %s : %ld", user, (long) pw->pw_uid);
    if (uid == pw->pw_uid)
    {
      LOG_MSG_WARNING("sympa-milter is already running as user %s", user);
    } else
    {
      if ((uid == 0) && (setreuid(pw->pw_uid, pw->pw_uid) < 0))
      {
        LOG_SYS_ERROR("Can't set process uid = %ld", (long) pw->pw_uid);
        return 1;
      }
    }
  } else
  {
    LOG_SYS_ERROR("Error getpwnam(%s)", user);
    return 1;
  }

  return res;
}

/* ************************************
 *                                    * 
 *                                    *
 ************************************ */

int
daemon_init()
{
  char               *user, *group;
  int                 fd;

  MESSAGE_INFO(9, " Let's daemonize sympa-milter...");
  switch (fork())
  {
    case 0:
      MESSAGE_INFO(11, " sympa-milter daemonized !");
      break;
    case -1:
      LOG_SYS_ERROR(" Error daemonizing sympa-milter ");
      exit(1);
      break;
    default:
      exit(0);
  }

  if (setpgid(0, 0) < 0)
    LOG_SYS_ERROR("Can't set process group leader");

  signal(SIGHUP, SIG_IGN);
  switch (fork())
  {
    case 0:
      break;
    case -1:
      LOG_SYS_ERROR(" Error daemonizing sympa-milter ");
      exit(1);
      break;
    default:
      exit(0);
  }

  set_log_output(TRUE, FALSE);

  umask(0000);

  if ((fd = open("/dev/null", O_RDONLY, 0)) < 0)
    LOG_SYS_ERROR("Can't open /dev/null read-only");
  if (dup2(fd, STDIN_FILENO) < 0)
    LOG_SYS_ERROR("Can't redirect stdin");
  close(fd);
  if ((fd = open("/dev/null", O_WRONLY, 0)) < 0)
    LOG_SYS_ERROR("Can't open /dev/null write-only");
  if (dup2(fd, STDOUT_FILENO) < 0)
    LOG_SYS_ERROR("Can't redirect stdout");
  if (dup2(fd, STDERR_FILENO) < 0)
    LOG_SYS_ERROR("Can't redirect stderr");
  close(fd);

  (void) set_uid_gid(sympa_user, sympa_group);

  if (chdir(sympa_spool_dir) != 0) {
    LOG_SYS_ERROR("Can't change current directory to %s", sympa_spool_dir);
    exit (EX_SOFTWARE);
  }
  return 0;
}
